"""Processor methods.
"""

import re


class SimpleCleaner(object):
    """Very simple data cleaner."""

    def __init__(self):
        """Just boot."""
        self.emoticons = set((
            '*)', '*-)', '8)', '8-)', '8-d', ":'''(", ":'(", ':(', ':)',
            ':-(', ':-)', ':-.', ':-/', ':-<', ':-d', ':-o', ':-p', ':-s',
            ':-[', ':-b', ':-c', ':-o', ':-p', ':-s', ':-|', ':/', ':3', ':>',
            ';d', ';p', ';x', 'x|', ':f',
            ':d', ':o', ':[', ':\\', ':]', ':^)', ':b', ':c', ':?',
            ':c)', ':o', ':o)', ':p', ':s', ':{', ':|', ':}', ";'(", ';)',
            ';-)', ';-]', ';D', ';]', ';^)', '<3', '=(', '=)', '=-d', '=/',
            '=d', '=]', '>.>', '>:)', '>:/', '>:d', '>:P', '>:[', '>:\\',
            '>:o', '>;]', 'x-d', 'xd', 'o.o', 'o_o', 'x-d', u'\xb0O\xb0',
            u'\xb0o\xb0', u'\u2665', u'\u2764', '^_^', '-_-'
        ))

    def clean(self, text):
        """Lower text and removes special stuff."""
        text = text.lower().replace('|', ' ')  # ask.fm specific
        text = ' '.join([re.sub('[^a-zA-Z0-9-_*?!]', ' ', word) if not
                         any([emo in word for emo in self.emoticons]) else
                         ' ' + word for word in text.split()])
        text = re.sub(r'\s{2,}', ' ', text)
        return text


class SocialCleaner(object):
    """Preprocess raw social media data.

    This class should handle all the preprocessing done for both the labels
    as well as the texts that are provided. Current implementations are the
    standard replacement of bbcode, emoticons and urls with their own
    __TOKENS__. These are placed in the `basic` preprocessing.
    """

    def __init__(self):
        """Record the applied methods."""
        self.applied = []

    def clean(self, text):
        """Clean according to ALL the preprocessors."""
        text = self.replace_url_email(text)
        text = self.find_emoticons(text)
        text = self.replace_bbcode_tags(text)
        return text

    def replace_bbcode_tags(self, text):
        """Replace BBCode tags.

        Replace all tags with [], which are included in the Netlog data,
        with a tag consisting of capital letters surrounded by underscores.
        Typography tags such as [b], [/b], [u], [/u], [i], [/i] (for bold,
        underlined and italics) are removed.

        The new tags are:
        - _PHOTO_, [photo]116157181[/photo]
        - _VIDEO_, [video]nl-9159440[/video]
        - _URL_, [url=http://www.adres.be/]Adres[/url] or [/url]
        - _EMOTICON_, [love], [@hug], [#clap_anim]

        Parameters
        -----
        text : string
            Input text.

        Returns
        -----
        text : string
            The text in which the Netlog tags with [] have been replaced.
        """
        self.applied.append('repl_bbcode')

        url = re.compile(r"(?:\[url=[^\]]+\][^\[]+)?\[\/url\]")
        photo = re.compile(r"\[photo\][^\]]+\[\/photo\]")
        video = re.compile(r"\[video\][^\[\]]+\[\/video\]")
        typo = re.compile(r"\[\/?[biu]\]")  # bold italics etc.
        emos = re.compile(r"\[[^\[\]]+\]")  # leftovers are probably emots

        text = url.sub('_URL_', text)
        text = photo.sub('_PHOTO_', text)
        text = video.sub('_VIDEO_', text)
        text = typo.sub('', text)
        text = emos.sub('_EMOTICON_', text)
        return text

    def replace_url_email(self, text, repl=('_URL_', '_EMAIL_')):
        """Replace URLs and e-mail addresses.

        Replace URLs with the tag _URL_
        Replace e-mail addresses with the tag _EMAIL_

        Parameters
        -----
        text : string
            Input text.

        Returns
        -----
        text : string
            The text in which the URLs and e-mail addresses have been replaced.
        """
        self.applied.append('repl_url')
        url = re.compile(r"(?:http[s]?://)www.(?:[a-zA-Z]|[0-9]|[$-_@.&+]|" +
                         r"[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+")
        email = re.compile(r"\S+\@[\w\.\-]+\.\w+")
        text = re.sub(url, repl[0], text)
        text = re.sub(email, repl[1], text)
        return text

    def find_emoticons(self, text, repl="_EMOTICON_"):
        """Replace or find emoticons in given text.

        Replace emoticons with a replacement string (default="_EMOTICON_").

        Emoticons can be western (and flipped) -- :), :p, :(, o:, x: -- or
        eastern ^_^.

        Parameters
        -----
        text : string
            Input text.

        repl_str : string, default "_EMOTICON_"
            String with which emoticons are replaced.

        Returns
        -----
        re.sub : string
            The text with the emoticons replaced by repl.
        re.findall : list
            If there is no specified replacement tag, will just return list of
            emoticons.
        """
        if repl:
            self.applied.append('repl_emoticons')
        emoticon_string = r'''
        (?:
          [<>]?
          [:;=8xX]                         # eyes L
          [\-o\*\']?                       # nose L
          [\)\]\(\[dDpPxXoOsS/\:\}\{@\|\\] # mouth L
          |
          [\)\]\(\[dDpPxXoOsS/\:\}\{@\|\\] # mouth R
          [\-o\*\']?                       # nose R
          [:;=8xX]                         # eyes R
          [<>]?
          |
          [\~<>]?                          # hands
          [\(] ?                           # body L
          [\-Oo~\^]                        # eyes L
          [\_\-\.]                         # mouth
          [\-Oo~\^]                        # eyes R
          [\)] ?                           # body R
          [\~<>]?                          # hands
        )'''
        pat = re.compile(emoticon_string, re.VERBOSE | re.I | re.UNICODE)
        return re.sub(pat, repl, text) if repl else re.findall(pat, text)


class Spacy(object):
    """Wrapper to spaCy.io. From their docs @ http://http://spacy.io/docs/.

    "spaCy consists of a vocabulary table that stores lexical types, a
    pipeline that produce annotations, and three classes to manipulate
    document, span and token data. The annotations are predicted using
    statistical models, according to specifications that follow common
    practice in the research community."

    spaCy is currently used in Omesa to provide the English part of the
    backbone. It's faster than CoreNLP, and Python <3. While spaCy can also
    extract things such as NER (it lacks sentiment and co-reference), this
    is currently not enabled for Omesa.
    """

    def __init__(self, raw=True):
        """Load up the spaCy pipeline."""
        from spacy.en import English
        self.spacy = English()
        self.raw = raw

    def parse(self, text):
        """Extract spaCy tags.

        Convert raw text instance into spaCy format. Currently only returns
        token, lemma, POS.

        Parameters
        -----
        text : string
            A raw string of characters.

        Returns
        -----
        instance : list
            The token, lemma, POS list that can be used in featurizers.
        """
        doc = self.spacy(text, tag=True, parse=True)
        return [[token.orth_, token.lemma_, token.pos_] for token in doc]


class Frog(object):
    """Wrapper to python-frog, loaded from LaMachine.

    Excerpt from the documentation @ http://ilk.uvt.nl/frog/:

    Frog is an integration of memory-based natural language processing (NLP)
    modules developed for Dutch. All NLP modules are based on Timbl, the
    Tilburg memory-based learning software package. Recently, a dependency
    parser, a base phrase chunker, and a named-entity recognizer module were
    added. Where possible, Frog makes use of multi-processor support to run
    subtasks in parallel.

    Frog is currently used in Omesa to provide the Dutch part of the backbone.
    As the other backbones, it currently only uses a subset of features. Full
    list of potential extractions (not enabled for Omesa) are:

    - Morphological segmentation (according to MBMA).
    - Confidence in the POS tag, a number between 0 and 1, representing the
      probability mass assigned to the best guess tag in the tag distribution.
    - Named entity type, identifying person (PER), organization (ORG), location
      (LOC), product (PRO), event (EVE), and miscellaneous (MISC), using a BIO
      (or IOB2) encoding.
    - Base (non-embedded) phrase chunk in BIO encoding.
    - Token number of head word in dependency graph (according to CSI-DP).
    - Type of dependency relation with head word.

    Notes
    -----
    General help: Florian Kunneman
    """

    def __init__(self, lmdir, sleep=False, raw=True):
        """Start the frog server if the sleep function isn't on."""
        if not sleep:
            import frog
            opts = frog.FrogOptions(parser=False, ner=False)
            self.frogger = frog.Frog(opts, lmdir + "LaMachine/lamachine/etc/"
                                     "frog/frog-twitter.cfg")
            self.raw = raw

    def parse(self, text):
        """Extract frog tags.

        Convert raw text instance into Frog format. Currently only returns
        token, lemma, POS.

        Parameters
        -----
        text : string
            A raw string of characters.

        Returns
        -----
        instance : list
            The token, lemma, POS list that can be used in featurizers.
        """
        doc = self.frogger.process(text)
        return [[token["text"], token["lemma"], token["pos"]] for token in doc]



"""LIME evaluation from Ribeiro, Singh, and Guestrin (2016)."""

import csv

import plotly.offline as py
import plotly.graph_objs as go

from lime.lime_text import ScikitClassifier, LimeTextExplainer


class LimeEval(object):
    """Local linear approximation of a model's behaviour.

    Lime is able to explain any black box text classifier, with two or more
    classes. All it requires is that the classifier implements a function that
    takes in raw text and outputs a probability for each class. Support for
    scikit-learn classifiers is built-in.

    On top of the author's code, this provides several functions to make it
    work with both the Omesa pipeline, as well as the front-end.

    Parameters
    ----------
    cls_ : class
        Scikit-learn classifier.

    vectorizer: class
        Can be either a Scikit-learn classifier or omesa.components.vectorizer.

    Attributes
    ----------
    c : class
        Initialized lime.ScikitlearnClassifier.

    names : array-type of strings
        List (or array) with string versions of labels (to be displayed).

    docs : list of strings
        List of input documents. These are automatically retrieved from the
        configuration["lime_data"] (if set). This process is handled when using
        LimeEval.load_omesa.

    Examples
    --------
    >>> grab = sklearn.datasets.fetch_20newsgroups
    >>> D, Di = grab(subset='train'), grab(subset='test')

    >>> vec = sklearn.feature_extraction.text.TfidfVectorizer(lowercase=False)
    >>> X = vec.fit_transform(D.data)
    >>> Xi = vec.transform(Di.data)

    >>> rf = sklearn.ensemble.RandomForestClassifier(n_estimators=500)
    >>> rf.fit(X, D.target)

    >>> import omesa.tools.lime_eval as le
    >>> le.LimeEval(rf, vec, class_names=list(set(Di.target)))
    >>> exps = le.explain(Di[:5])
    >>> graph_to_file(exps, '/some/file/wherever')

    Notes
    -----
    Package from: https://github.com/marcotcr/lime.
    """

    def __init__(self, classifier=None, vectorizer=None, class_names=None,
                 docs=None):
        """Start lime classifier, set label and empty doc placeholder."""
        self.c = ScikitClassifier(classifier, vectorizer)
        self.names = class_names
        self.docs = [] if not docs else docs
        self.multi = 3 if len(class_names) > 2 else None

    def explain(self, docs):
        """Generate LIME Explanations for list of docs.

        Takes as input a list of strings that make up the documents where LIME
        should be applied to. Returns Explanation class instances.

        Parameters
        ----------
        docs : list of strings
            List of input documents.

        Returns
        -------
        exps : list of classes
            For each input document, an Explanation class object on which for
            example the .to_list, to_notebook etc functions can be called on.
        """
        explainer = LimeTextExplainer(class_names=self.names)
        exps = []
        for doc in docs:  # NOTE: this might have messed up in a generator
            exp = explainer.explain_instance(doc, self.c.predict_proba,
                                             top_labels=self.multi)
            exps.append(exp)
        return exps

    def load_omesa(self, lime_repr):
        """Special LIME loader for Omesa pipelines.

        Tries to find a path location to extract example documents from, and
        use the indices set in the omesa.containers object. This requires a
        __dict__ representation of the reader class to have self.path (full
        system path of a file), and self.idx (list of integers with
        [text_index, label_index, etc.]). Advisable to only be used icw Omesa.

        Parameters
        ----------
        lime_repr : ...
            ...

        Returns
        -------
        docs : list of strings
            The top 5 (assuming it has a header) documents from a
            omesa.containers object.
        """
        if isinstance(lime_repr, dict):
            reader = csv.reader(open(lime_repr['path']), quotechar='"')
            ti = lime_repr['idx'][0]
            for i, row in enumerate(reader):
                if lime_repr.get('no_header') and not i:
                    self.docs.append(row[ti])
                elif i:
                    self.docs.append(row[ti])
                if i == 5:
                    break
        else:
            self.docs = lime_repr
        return self.explain(self.docs)

    @staticmethod
    def graph_to_file(exps, loc):
        """Dump LIME experiments with .to_html.

        This is the native way of graphing using LIME, and uses d3.js. The
        files are generally lighter (1 vs 3 MB) than using Omesa (uses plotly).
        However, as they would have to be embedded in the front-end using an
        iframe, the style cannot be changed. As such, this function is ommitted
        when using the Omesa front-end.

        Parameters
        ----------
        exps : list of classes
            An Explanation class for each document.

        loc : str
            The location where to save. If this is used in bottle, should be
            specifically set to `None` to save it to ./static.

        Returns
        -------
        f_names : list of strings
            List of pointers to the file locations where the graphs have been
            stored.
        """
        f_names = []
        for i, exp in enumerate(exps):
            if not loc:
                loc = '/tmp/plot/'
            with open(loc + 'lime_' + str(i) + '.html', 'w') as f:
                html_str = exp.as_html()
                f.write(html_str)
                f_names.append(f.name[1:])
        return f_names

    @staticmethod
    def save_graph(data, layout):
        """Quick binder to dump plotly data and layout."""
        fig = go.Figure(data=data, layout=layout)
        return py.plot(fig, output_type='div', auto_open=False,
                       show_link=False, include_plotlyjs=False)

    def prob_graph(self, i, prob, cln, cols):
        """Output LIME class probability graph. Works with 'graphs' method."""
        data = [go.Bar(x=list(prob), y=cln,
                       marker=dict(color=cols),
                       orientation='h')]
        layout = go.Layout(margin=go.Margin(l=100, r=0, b=0, t=0, pad=0))
        return self.save_graph(data, layout)

    def weight_graph(self, i, expl, cols):
        """Output LIME weight graph. Works with 'graphs' method."""
        # FIXME: colours are binary only
        data = [go.Bar(x=[float(val) for word, val in expl],
                       y=[word for word, val in expl],
                       marker=dict(color=[cols[0] if val < 0 else cols[1]
                                          for word, val in expl]),
                       orientation='h')]
        layout = go.Layout(margin=go.Margin(l=100, r=0, b=0, t=0, pad=0))
        return self.save_graph(data, layout)

    def tag_text(self, i, expl, cols):
        """Highlight LIME top-word in text. Works with 'graphs' method."""
        # FIXME: replace special chars with space and replace on token
        repl = [(word, ('__LIMENEG__' if val < 0 else '__LIMEPOS__') +
                 word + '</span></b>') for word, val in expl]
        doc = str(self.docs[i]).replace('"', '')
        for y in repl:
            doc = doc.replace(*y)
        # these are split up in tokens so that f.e. '1' doesn't screw it up
        doc = doc.replace(
            '__LIMENEG__', '<b><span style="color:{0}">'.format(cols[0]))
        doc = doc.replace(
            '__LIMEPOS__', '<b><span style="color:{0}">'.format(cols[1]))
        return doc

    def unwind(self, exp, comp=False):
        """Unwind LIME experiment in its results."""
        if not comp:
            expl = exp.as_list()
            prb = exp.predict_proba
            cln = exp.class_names
        else:
            expl, prb, cln = exp['expl'], exp['prb'], exp['cln']
        return expl, prb, cln

    def graphs(self, exps, comp=False):
        """Convert exps list to graph locations and annotated text."""
        import colorlover as cl
        order = []
        for i, exp in enumerate(exps):
            expl, prb, cln = self.unwind(exp, comp)
            try:
                ncol = cl.scales[str(len(cln))]['qual']['Set2']
            except KeyError:
                ncol = cl.scales[str(len(cln) + 1)]['qual']['Set2']
            cols = cl.to_rgb(ncol)
            order.append([self.prob_graph(i, prb, cln, cols),
                          self.weight_graph(i, expl, cols),
                          self.tag_text(i, expl, cols)])
        return order

    def to_web(self, tab):
        xps = tab.get('lime_data_comp')
        if not xps:
            xps = self.load_omesa(tab['lime_data_repr'])
        else:
            self.docs = tab.get('lime_data')
        return [x for x in self.graphs(xps, comp=isinstance(xps[0], dict))] \
            if xps else ["Model not probability-based."]



"""Vectorizer and optimization.
"""

from operator import itemgetter
from multiprocessing import Pool

import numpy as np

from sklearn import pipeline
from sklearn.preprocessing import LabelEncoder
from sklearn.feature_extraction import DictVectorizer
from sklearn.grid_search import GridSearchCV
from sklearn.svm import SVC

from .featurizer import Featurizer
from .containers import Pipe, _chain


class Vectorizer(object):
    """Small text mining vectorizer.

    The purpose of this class is to provide a small set of methods that can
    operate on the data provided in the Experiment class. It can load data
    from an iterator or .csv, and guides that data along a set of modules such
    as the feature extraction, tf*idf function, SVD, etc. It can be controlled
    through a settings dict that is provided in conf.

    Parameters
    ----------
    conf : dict
        Configuration dictionary passed to the experiment class.

    Attributes
    ----------
    conf : dict
        Configuration dictionary passed to the experiment class.

    featurizer : class
        Environment class (might be replace by Featurizer) from Omesa.

    hasher : class
        DictVectorizer class from sklearn.

    normalizers : class
        TfidfTransformer class from sklearn.

    decomposers : class
        TruncatedSVD class from sklearn.
    """

    def __init__(self, conf=None, featurizer=None):
        """Start pipeline modules."""
        if conf:
            self.conf = conf
        if not featurizer:
            self.featurizer = Featurizer(conf['features'],
                                         conf.get('preprocessor'),
                                         conf.get('parser'))

        self.hasher = DictVectorizer()
        self.encoder = LabelEncoder()

    def _vectorize(self, data, func):
        """Send the data through all applicable steps to vectorize."""
        if isinstance(data, list) and hasattr(data[0], 'source'):
            data = _chain(data)

        p = Pool(processes=None)
        D, y = zip(*p.map(self.featurizer.transform, data))
        p.close()
        p.join()
        del p

        # NOTE: these _can't_ be put in p.map because `fit` overwrites in iter
        X = getattr(self.hasher, func)(D)
        y = getattr(self.encoder, func)(y) if len(set(y)) != 1 else ''

        if len(y):
            return X, y
        else:
            return X

    def fit_transform(self, data):
        """Adhere to sklearn API."""
        return self._vectorize(data, func='fit_transform')

    def transform(self, data):
        """Adhere to sklearn API."""
        return self._vectorize(data, func='transform')


class Optimizer(object):
    """Current placeholder for grid methods. Should be fleshed out.

    Parameters
    ----------
    classifiers : dict, optional, default None
        Dictionary where the key is a initiated model class, and the values
        are a dictionary with parameter settings in a (string-array) format,
        same as used in the scikit-learn pipeline. So for example, we provide:
        {LinearSVC(class_weight='balanced'): {'C': np.logspace(-3, 2, 6)}}.
        Note that pipeline requires some namespace (like clf__C), but the class
        handles that already.

    conf : dict, optinal, default None
        Configuration dictionary used by the Experiment class wrapper.
    """

    def __init__(self, conf=None, classifiers=None, scoring='f1_micro'):
        """Initialize optimizer with classifier dict and scoring, or conf."""
        std_clf = [Pipe('clf', SVC(kernel='linear'),
                        parameters={'C': np.logspace(-2.0, 1.0, 10)})]

        if not classifiers:
            classifiers = std_clf
        if not conf.get('classifiers'):
            conf['classifiers'] = std_clf

        self.scores = {}
        self.met = conf.get('scoring', scoring)
        self.conf = conf if conf else classifiers

    def best_model(self):
        """Choose best parameters of trained classifiers."""
        score_sum, highest_score = {}, 0
        for scores, estim in self.scores.values():
            best = sorted(scores, key=itemgetter(1), reverse=True)[0]
            score = best.mean_validation_score
            score_sum[score] = estim
            if score > highest_score:
                highest_score = score

        print("\n #--------- Grid Results -----------")
        print("\n Best scores: {0}".format(
            {str(dict(y.steps)['clf'].__class__.__name__): round(x, 3)
             for x, y in score_sum.items()}))

        return highest_score, score_sum[highest_score]

    def choose_classifier(self, X, y, seed):
        """Choose a classifier based on settings."""
        clfs, pipes = [], []

        for pipe in self.conf['pipeline']:
            pipe.check(seed)
            if pipe.idf == 'clf':
                clfs.append(pipe)
            else:
                pipes.append(pipe)

        for i, clf in enumerate(clfs):
            grid = {pipe.idf + '__' + k: v for pipe in pipes
                    for k, v in pipe.parameters.items()}
            grid.update({'clf__' + k: v for k, v in clf.parameters.items()})

            print("\n #-------- Classifier {0} ----------".format(i + 1))
            print("\n", "Clf: ", str(clf.skobj))
            print("\n", "Grid: ", grid)

            grid = GridSearchCV(
                pipeline.Pipeline([(pipe.idf, pipe.skobj) for pipe in pipes] +
                                  [('clf', clf.skobj)]),
                scoring=self.met, param_grid=grid,
                n_jobs=self.conf.get('n_jobs', -1))

            print("\n Starting Grid Search...")
            grid.fit(X, y)
            print(" done!")

            self.scores[clf] = (grid.grid_scores_, grid.best_estimator_)

        score, clf = self.best_model()
        self.scores['best'] = score

        return clf



"""Classes to log and display pipeline information.
"""

from collections import OrderedDict


class _Logger(object):
    """Provides feedback to the user and can store settings in a log file.

    Class holds a log string that can be formatted according to the used
    components and is used to list settings that are provided to the
    experiment. Aside from flat printing, it can iteratively store certain
    lines that are reused (such as loading in multiple datasets). The save
    function makes sure the self.log items are saved according to their logical
    order.

    Parameters
    ----------
    fn : str
        File name of the logfile (and the experiment).

    Attributes
    ----------
    fn : str
        File name.

    log : dict
        Keys are short names for each step, values are strings with .format
        placeholders. Formatting is sort of handled by the strings as well.

    buffer : list
        Used to stack lines in a loop that can be written to the log line once
        the loop has been completed.
    """

    def __init__(self, file_name):
        """Set log dict. Empty buffer."""
        self.file_name = file_name + '.log'
        self.log = {
            'head':
                "\n---- Omesa ---- \n\n Config: \n" +
                "\t {0} \n\tname: {1} \n\tseed: {2} " +
                "\n\t \n",
            # 'read':
            #    "\n Reading from {0}... Acquired {1} from data.\n ",
            'sparse':
                "\n Sparse {0} shape: {1}",
            'svd':
                "\n Fitting SVD with {0} components...",
            'rep':
                "\n\n---- {0} Results ---- \n" +
                "\n Distribution: {1}" +
                "\n Accuracy @ baseline: \t {2}" +
                "\n Reporting on label {3}",
            'grid':
                "\n Model with rank: {0} " +
                "\n Mean validation score: {1:.3f} (std: {2:.3f}) " +
                "\n Parameters: {3} \n",
            'tfcv':
                "\n Tf-CV Result: {0}",
            'f1sc':
                "\n F1 Result: {0}",
            'cr_train':
                "\n Performance on train set: \n\n{0}",
            'cr_test':
                "\n Performance on test set: \n\n{0}",
            'slice':
                "\n\n # ---------- Grid slice {0} ------------\n"
        }
        self.buffer = []

    def head(self, features, name, seed):
        self.post('head', ('\n'.join([str(c) for c in features]), name, seed))

    def data(self, dtype, dset, data, dump=False):
        self.loop(dtype, (dset, data.shape))
        if dump:
            self.dump(dtype)

    def report(self, tt, yi, res, av, metrics, labs=None):
        self.post('cr_' + tt, (
            metrics.classification_report(yi, res, target_names=labs), ))
        return {'y': yi, 'res': res,
                'score': metrics.f1_score(yi, res, average=av)}

    def slice(self, slicenr):
        self.post('slice', slicenr)

    @staticmethod
    def echo(*args):
        """Replacement for a print statement. Legacy function."""
        message = ' '.join([str(x) for x in args])
        print(message)

    def loop(self, key, value):
        """Print and store line to buffer."""
        line = self.log[key].format(*value)
        print(line)
        self.buffer.append(line)

    def dump(self, key):
        """Dump buffer to log."""
        self.log[key] = ''.join(self.buffer)
        self.buffer = []

    def post(self, key, value):
        """Print and store line to log."""
        line = self.log[key].format(*value)
        print(line)
        self.log[key] = line

    def save(self):
        """Save log."""
        with open(self.file_name, 'w') as f:
            o = ['head', 'sparse', 'svd', 'rep', 'grid', 'tfcv', 'f1sc',
                 'cr_train', 'cr_test', 'slice']
            f.write(' '.join(
                [v for v in OrderedDict(
                    sorted(self.log.items(), key=lambda i: o.index(i[0]))
                    ).values()]))



"""Converts sklearn-styled docstrings to markdown."""

from sys import argv

# Author:       Chris Emmery
# License:      MIT

ERR = '''
- """ used for non-docstring strings.
- Class or function Documentation has newline after opening """.
'''


class Doc2Markdown(object):
    """Documentation to Markdown."""

    def __init__(self, file_in, file_out):
        """Handle in/out to readers and parsers."""
        self.file_in = file_in
        self.file_out = file_out
        self.markdown = ''

        self.md_table = '''

| {0}    | Type             | Doc             |
|:-------|:-----------------|:----------------|
        '''
        self.md_table_row = '''| {0} | {1} | {2} |
        '''

        self.md_table_alt = '''

| {0}    | Doc             |
|:-------|:----------------|
        '''
        self.md_table_row_alt = '''| {0} | {1} |
        '''

        self.md_code = '``` python \n {0} \n```\n\n'

        self.read()

    def read(self):
        """Open pytyhon code file, spit out markdown file."""
        with open(self.file_in, 'r') as file_in:
            file_open = file_in.read()
            classes = file_open.split('\nclass ')
            module_doc = classes.pop(0)  # top of the file docstring
            self.markdown = [self.md_title(module_doc)]
            self.handle_classes(classes)
        with open(self.file_out, 'w') as file_out:
            file_out.write('\n'.join(self.markdown))

    def handle_classes(self, classes):
        """Given class docs only, write out their docs and functions."""
        for clss in classes:
            clss = " class " + clss  # got cut-off in split
            clss_str = clss.split('"""\n')[:-1]
            # process all the class information
            self.markdown.append(
                self.md_class_doc(self.split_doc(clss_str.pop(0))))
            # replace object part with __init__ params
            self.markdown[1] = \
                self.md_class_func(self.markdown[1], clss_str)
            # start handling def
            self.markdown.append(
                self.md_funcs_doc([self.split_doc(func) for func in clss_str]))

    @staticmethod
    def split_doc(doc):
        """Split a python file on docstring."""
        lines = [x for x in doc.split('\n') if x]
        name, det = '', False
        # for i, l in enumerate(lines):
        #     print("{0}: ".format(i+1), l)
        # skip lines until class/def, end if :
        while ':' not in name:
            try:
                if 'class ' in lines[0] or 'def ' in lines[0]:
                    det = True
            except IndexError:
                exit("Malformed code. \n\nSome common errors: {0}".format(ERR))
            line = lines.pop(0)
            if det:
                name += line

        parts = {'name': name}
        buffer_to = 'title'
        # if ----- is found, set title to above, else buffer to name
        for i in range(0, len(lines)):
            line = lines[i]
            if '---' in line:
                buffer_to = lines[i-1]
                continue
            if not parts.get(buffer_to):
                parts[buffer_to] = [buffer_to]
            if parts.get(buffer_to) and line:
                parts[buffer_to].append(line)
        # if line is shorter than 3, likely that there is no previous section
        return {k.replace('  ', ''):
                (v[:-1] if len(v) > 2 else v)[1:] for k, v in parts.items()}

    @staticmethod
    def md_title(title_part, class_line=False):
        """Handle title and text parts for main and class-specific."""
        buff, start, second = [], False, False
        # pretty convoluted function just to append # to first line and clean
        for line in title_part.split('\n'):
            line = line.replace('  ', '')
            if line.startswith('"""') and not start:
                title = '# ' + line.replace('"""', '')[:-1]
                if class_line:
                    title = title[2:] + '.'
                buff.append(title)
                start = True
            elif line.startswith('"""') and start:
                start = False
                return '\n'.join(buff)
            elif start:
                if not second and not line.startswith('\n'):
                    line = "\n" + line
                    second = True
                buff.append(line)
        return '\n'.join(buff)

    def md_par_att(self, doc, name):
        """Place parameters and attributes in a table overview."""
        head, buff, lines = '', (), []
        table = self.md_table.format(name)
        # given var : type \n description, splits these up into 3 cells
        if doc:
            for row in doc:
                if ' : ' in row and not buff:
                    head = row.split(' : ')
                    for part in head:
                        buff += (part.replace('  ', ''), )
                elif ' : ' in row and buff:
                    buff += (''.join(lines), )
                    table += self.md_table_row.format(*buff)
                    buff, lines = (), []
                elif buff:
                    lines.append(row.replace('\n', ' '))
            if buff:
                buff += (''.join(lines).replace('  ', ''), )
                table += self.md_table_row.format(*buff)
            return table
        else:
            return ''

    def md_examples(self, doc):
        """Section text and code in the example part of the doc."""
        head = '\n\n------- \n\n##Examples\n\n{0}'
        order = []
        text, code = '', ''
        # if we find some code related beginnings (>>>, ...) buffer to code,
        # else we regard it as text. Record order so that multiple blocks of
        # text and code are possible.
        if doc:
            for row in doc:
                if '>>>' in row or '...' in row:
                    if text:
                        order.append(text)
                        text = ''
                    row = row.replace('    >>>', '>>>')
                    row = row.replace('    ...', '...')
                    code += row + '\n'
                else:
                    if code:
                        order.append(self.md_code.format('\n' + code))
                        code = ''
                    row = row.replace('  ', '')
                    text += row + '\n'
            if text or code:
                order.append(text)
                order.append(self.md_code.format('\n' + code))
            return head.format('\n'.join(order).replace('.\n', '.\n\n'))
        else:
            return ''

    def md_class_func(self, doc, clss):
        """Replace class ...(object) with __init__ arguments."""
        init_doc = self.split_doc(clss.pop(0))['name']
        init_doc = init_doc.replace('  ', '')
        init_doc = init_doc.replace(' def __init__(self, ', '(')
        return doc.replace('(object)', init_doc)

    def md_funcs_doc(self, func_doc):
        """Merge all function elements in one string."""
        mark_doc, mark_head = '', '\n--------- \n\n## Methods \n\n {0} \n {1}'
        func_table = self.md_table_alt.format('Function')
        for func in func_doc:
            name = func['name'].replace('(', ' (').split()[1]
            title = self.md_title('\n'.join(func['title']), class_line=True)
            func_table += self.md_table_row_alt.format(name,
                                                       title.split('\n')[0])
            mark_doc += '\n\n### ' + name + '\n\n'

            cod = self.md_code.format(func['name'].replace('def ', ''))
            cod = cod.replace('self', '').replace('(, ', '(')

            mark_doc += cod
            mark_doc += '\n' + title
            mark_doc += self.md_par_att(func.get('Parameters'), 'Parameters')
            mark_doc += self.md_par_att(func.get('Returns'), 'Returns')
        return mark_head.format(func_table, mark_doc)

    def md_class_doc(self, class_doc):
        """Merge all class section elements in one string."""
        mark_doc = ''
        mark_doc += '\n\n# {0} \n\n'.format(
            class_doc['name'].replace('(', ' (').split()[1])
        mark_doc += self.md_code.format(class_doc['name'])
        mark_doc += self.md_title('\n'.join(class_doc['title']),
                                  class_line=True)
        mark_doc += self.md_par_att(class_doc.get('Parameters'), 'Parameters')
        mark_doc += self.md_par_att(class_doc.get('Attributes'), 'Attributes')
        mark_doc += self.md_examples(class_doc.get('Examples'))
        return mark_doc

if __name__ == '__main__':
    D2M = Doc2Markdown(argv[1], argv[2])



"""Data handling containers.
"""

# pylint:       disable=W0234

import csv
import json
import pickle
import sys

from os import getcwd
from types import GeneratorType

from .tools import serialize_sk as sr

try:
    from .database import Database, Configuration, Vectorizer, \
                          Classifier, Results, Table
except ImportError as e:
    print(e)
    print("Database could not be loaded, functionality disabled.")


def _chain(data):
    """Chain containers."""
    for container in data:
        for entry in container:
            yield entry



class Pipeline(object):
    """Shell for experiment pipeline storing and handling.

    Parameters
    ----------
    exp : class, optional, default None
        Instance of Experiment with fitted components. If not supplied, name
        and store should be set.

    name : str, optional, default None
        Name that the pipeline should be saved/loaded under/from.

    store : tuple, optional, default None
        Tuple with storage options, can be "json" (json serialization),
        or "db" (for database storage, requires blitzdb).
    """

    def __init__(self, exp=None, name=None, store=None):
        """Set the pipeline for transformation and clf for classification."""
        if not exp:
            assert name

        self.vec = exp.vec if exp else None
        self.cnf = exp.vec.conf if exp else None
        self.clf = exp.clf if exp else None
        self.res = exp.res if exp else None

        self.hook = self.vec.conf['name'] if not name else name
        self.storage = self.vec.conf['save'] if not store else store

        if 'db' in self.storage:
            self.db = Database()

    def _convert_data(self, tab, n):
        """Split data entries and leave empty if not present."""
        did = n + '_data'
        row = self.cnf.get(did)

        if not row:
            val = [[]] * 3
            val[0] = ['split'] if n == 'test' else ['-']
        elif isinstance(row, GeneratorType):
            val = [[]] * 3
            val[0] = ['loader'] if n != 'lime' else row
        elif isinstance(row, list) and isinstance(row[0], str):
            val = [[]] * 3
            val[2] = [x for x in row]
            val[0] = val[2]  # FIXME: wtf
        elif isinstance(row, list) and hasattr(row[0], 'source'):
            val = ([x.source for x in row], [x.path for x in row],
                   [x.__dict__ for x in row])
        elif n != 'lime':
            val = ([row.source], [row.path], [row.__dict__])
        else:
            # FIXME: weird that we store it in data here and repr if no csv
            val = ([x[0] for x in row], row.path, row.__dict__)

        tab.update({did: val[0], did + '_path': val[1], did + '_repr': val[2]})
        return tab

    def _calc_lime(self, tab):
        """Calculate lime information based on converted data entries."""
        from .tools import lime_eval as le
        labs = self.vec.encoder.classes_
        lime = le.LimeEval(self.clf, self.vec, labs)
        exps = lime.load_omesa(tab['lime_data_repr'])
        lime_list = []
        for exp in exps:
            expl, prb, cln = lime.unwind(exp)
            lime_list.append({'expl': expl, 'prb': prb, 'cln': cln})
        tab.update({'lime_data_comp': lime_list})
        return tab

    def _make_tab(self):
        """Tabular level experiment representation.

        Generates a table-level representation of an experiment. This stores
        JSON native information ONLY, and is used for the experiment table in
        the front-end, as deserializing a lot of experiments will be expensive
        in terms of loading times.
        """
        tab = {'project': self.cnf.get('project', '-'), 'name': self.hook,
               'clf': self.clf.__dict__['steps'][0][1].__class__.__name__,
               'clf_full': str(self.clf.__dict__['steps'][0][1])}

        for n in ('train', 'test', 'lime'):
            tab = self._convert_data(tab, n)
            if n == 'lime' and not self.cnf.get('lime_protect'):
                tab = self._calc_lime(tab)

        tab.update({'features': ','.join([x.__str__() for x in
                                          self.vec.featurizer.helpers]),
                    'test_score': self.res['test']['score'],
                    'dur': self.res['dur']})
        return tab

    def save(self):
        """Save experiment and classifier in format specified."""
        print(" Saving experiment...")
        tab = self._make_tab()
        fl = self.hook

        try:  # purge double vec instance
            self.vec.conf = self.hook
        except AttributeError:  # when loading
            pass

        if any([x in self.storage for x in ('json', 'pickle')]):
            top = {'name': self.hook, 'cnf': self.cnf, 'vec': self.vec,
                   'clf': self.clf, 'res': self.res, 'tab': tab}
        if 'json' in self.storage:
            sr.encode(top, open(self.hook + '.json', 'w'))
        if 'pickle' in self.storage:
            for t in ('train', 'test', 'lime'):
                c = top['conf']['{0}_data'].format(t)
                c = '' if isinstance(c, GeneratorType) else c
            pickle.dump(top, open(fl + '.pickle', 'wb'))
        if 'db' in self.storage:
            top = {Configuration: self.cnf, Vectorizer: self.vec,
                   Classifier: self.clf, Results: self.res, Table: tab}
            for doc, bind in top.items():
                js = json.loads(sr.encode(bind))
                js['name'] = self.hook
                self.db.save(doc(js))

    def load(self):
        """Load experiment and classifier from source specified."""
        if 'json' in self.storage:
            # FIXME: try to recursively solve imports if that works
            mod = sr.decode(open(self.hook + '.json'))
        if 'pickle' in self.storage:
            mod = pickle.load(open(self.hook + '.pickle', 'rb'))
        if 'db' in self.storage:
            mod = {'clf': Classifier, 'vec': Vectorizer, 'res': Results,
                   'cnf': Configuration}
            for k, doc in mod.items():
                mod[k] = self.db.get_component(doc, self.hook)
        self.clf = mod['clf']
        self.vec = mod['vec']
        if 'db' in self.storage:
            self.vec.conf = mod['cnf']
        self.res = mod['res']

    def classify(self, data, best_only=False):
        """Given instance(s) return list with (label, probabilities).

        Parameters
        ----------
        data : value or list
            Can be one or multiple data instances (strings for example).

        best_only : bool, optional, default False
            If set to True, returns probabilties for highest only.

        Returns
        -------
        self.clf.predict : list
            List with one or more tuples with (label, array of probabilities).
        """
        if isinstance(data, str):
            data = [data]
        X = self.vec.transform(data)
        try:
            prob_d = [{self.vec.encoder.inverse_transform(i): p
                       for i, p in enumerate(pl)}
                      for pl in self.clf.predict_proba(X)]
            if best_only:
                return sorted(prob_d.items(), key=lambda x: x[1])[-1]
            return prob_d
        except AttributeError:
            pass
        return self.vec.encoder.inverse_transform(self.clf.predict(X))


class CSV(object):
    """Quick and dirty csv loader.

    Parameters
    ----------
    text : integer or string
        Index integer of the .csv where the text is located. If string is
        provided instead, will look for its index in the header.

    label : integer or string
        Index integer of the .csv where the label is located. If string is
        provided instead, will look for its index in the header.

    parse : integer, optional, default None
        Index integer of the .csv where the annotations are provided. If string
        is provided instead, will look for its index in the header. Currently
        it assumes that these are per instance a list of, for every word,
        (token, lemma, POS). Frog and spaCy are implemented to provide these
        for you.

    features : list of integers or strings, optional, default None
        If you have columns in your .csv that should serve as features
        (meta-data) for example, you can add a multitude of their indices in
        this setting in integer format. If string is provided instead, will
        look for its index in the header.

    no_header : boolean, optional, default False
        If the file has no header, and integer values are provided as column
        indices, set to True if it has to be included.

    selection : dict, optional, default None
        A dict of {label: integer amount} pairs. With this, you can somewhat
        control how many of a certain label you wish to include in loading your
        data (due to memory constraints for example). If you want all, you can
        just put -1.
    """

    def __init__(self, csv_dir, data, label, parse=None, features=None,
                 no_header=False, selection=None):
        """Set configuration and label handler."""
        csv.field_size_limit(sys.maxsize)
        self.source = csv_dir
        self.path = getcwd() + ("/" if not csv_dir.startswith('/') else '') + \
            csv_dir
        self.file = csv.reader(open(csv_dir, 'r'))
        self.no_header = no_header

        if not no_header or isinstance(data, str):
            head = self.file.__next__()

        if isinstance(features, (int, type(None))):
            features = [features]

        self.idx = list(filter(None.__ne__, [data, label, parse] + features))
        self.selection = {} if not selection else selection

        if isinstance(data, str):
            self.idx = [head.index(ind) for ind in self.idx]

    def __iter__(self):
        """Standard iter method."""
        return self

    def __next__(self):
        """Iterate through csv file."""
        row = self.file.__next__()
        if self.selection.get(self.idx[0]):
            if self.selection[self.idx[0]]:
                self.selection[self.idx[0]] -= 1
                return tuple(row[i] for i in self.idx)
        else:
            return tuple(row[i] for i in self.idx)


class Pipe(object):
    """Pipe in pipeline wrapper.

    Used to cleanly handle pipeline components that adhere to the scikit-learn
    API, meaning they have a fit/transform method.

    Parameters
    ----------
    idf : string
        Name representation used in the pipeline. Only 'clf' should be used to
        selection of multiple classifiers.

    skobj : class
        Classifier, normalizer, or decompostion class that adheres to the
        scikit-learn API.

    parameters : dict
        With {'parameter name': values}. If an iterator or array-like object is
        provided as parameter, these will by default be used as combinations to
        apply grid search on.
    """

    def __init__(self, idf, skobj, parameters=False):
        """Pipe initialization."""
        self.idf = idf
        self.skobj = skobj
        self.parameters = parameters if parameters else {}

    def check(self, seed):
        """Check if correct params are set for objects."""
        try:
            self.skobj.copy = False
        except AttributeError:
            pass

        try:
            self.skobj.probability = True
            self.skobj.random_state = seed
        except AttributeError:
            pass



"""Blitzdb database declarations."""

from blitzdb import Document
from blitzdb import FileBackend
from os.path import expanduser
from .tools import serialize_sk as sr


class Configuration(Document):
    """Blitzdb configuration placeholder."""

    class Meta(Document.Meta):
        primary_key = 'name'


class Vectorizer(Document):
    """Blitzdb vectorizer placeholder."""

    class Meta(Document.Meta):
        primary_key = 'name'


class Classifier(Document):
    """Blitzdb classifier placeholder."""

    class Meta(Document.Meta):
        primary_key = 'name'


class Results(Document):
    """Blitzdb results placeholder."""

    class Meta(Document.Meta):
        primary_key = 'name'


class Table(Document):
    """Blitzdb table placeholder."""

    class Meta(Document.Meta):
        primary_key = 'name'


class Database(object):
    """Blitzdb database."""

    def __init__(self):
        """Load backend."""
        self.db = FileBackend(expanduser("~/.omesa/db"))

    def _query(self, f, q):
        try:
            out = f(*q)
        except KeyError:
            self.db = FileBackend(expanduser("~/.omesa/db"))
            f = self.db.filter
            out = f(*q)
        return out

    def save(self, doc):
        """Save document do db."""
        self.db.save(doc)
        self.db.commit()

    def fetch(self, doc, q):
        """Filter and return first entry."""
        try:
            return self._query(self.db.filter, (doc, q))[0]
        except IndexError:
            print(str(doc), str(q))
            print("File does not exist.")

    def get_component(self, doc, name):
        # FIXME: see if returning non-decoded is relevant for anything
        try:
            return sr.decode(dict(self._query(
                self.db.filter, (doc, {'name': name}))[0]))
        except IndexError:
            print(str(doc), {'name': name})
            print("File does not exist.")

    def getall(self, doc):
        """Returns all entries in db."""
        return [d for d in self._query(self.db.filter, (doc, {}))]



# -*- coding: utf-8 -*-

"""Text feature extraction module.

This module contains several helper classes for extracting textual features
used in Text Mining applications, partly based on instances parsed with parse.
It also includes a wrapper class to cleverly hanlde this within the Omesa
framework.
"""

import re
import os
import json
import pickle
from collections import OrderedDict, Counter
from urllib.parse import urlencode

import numpy as np


class Featurizer(object):
    """Wrapper for looping feature extractors in fit and transform operations.

    Calls helper classes which extract different features from text data. Given
    a list of initialized feature extractor classes, correctly streams or dumps
    instances along these classes. Also provides an interface to fit and
    transform methods.

    Parameters
    ----------
    features : list
        List of initialized feature extractor classes. The classes can be
        found within this module.

    Attributes
    ----------
    helper : list of classes
        Store for the provided features.

    X : list of lists of shape [n_samples, n_words]
        All data instances used by space_based featurizer helpers.

    Y : list of labels
        Labels for X.

    Examples
    --------
    Note: this is just for local use only.

    During training with a full space and a generator:
    >>> loader = reader.load  # assumes that this is a generator
    >>> features = [Ngrams(level='char', n_list=[1,2])]
    >>> ftr = _Featurizer(features)
    >>> ftr.fit(loader())
    >>> X, Y = ftr.transform(loader()), ftr.labels

    During testing with only one instance:
    >>> new_data = 'this is some string to test'
    >>> tex, tey = ftr.transform(new_data), ftr.labels

    Notes
    -----
    For an explanation regarding the parse features, please refer either to
    utils.parse.extract_tags or http://ilk.uvt.nl/parse/.
    """

    def __init__(self, features, preprocessor=False, parser=False):
        """Initialize the wrapper and set the provided features to a var."""
        self.helpers = features
        self.preprocessor = preprocessor
        self.parser = parser
        self.head = []

    def transform(self, instance):
        """Call all the helpers to extract features.

        Parameters
        ----------
        instance : tuple
            Containing at least (raw) and optionally (parse, meta).

        Returns
        -------
        v : dict
            Feature vector where key, value = feature, value.
        label : str
        """
        if isinstance(instance, str):
            instance = tuple([instance])
        raw, label, parse, meta = instance + (None,) * (4 - (len(instance)))

        text = self.preprocessor.clean(raw) if self.preprocessor else raw
        if not parse and self.parser:
            parse = self.parser.parse(raw if self.parser.raw else text)

        v = {}
        for helper in self.helpers:
            v.update(helper.transform(text, parse))
        if meta:
            for name, value in meta:
                v.update({"meta_" + name: value})

        return v, label


class Ngrams(object):
    """Calculate n-gram frequencies.

    Can either be applied on token, POS or character level. The transform
    method dumps a feature dictionary that can be used for feature hashing.

    Parameters
    ----------
    n_list : list of integers
        Amount of grams that have to be extracted, can be multiple. Say that
        uni and bigrams have to be extracted, n_list has to be [1, 2].

    max_feats : integers
        Limits how many features will be generated.

    Examples
    --------
    Token-level uni and bigrams with a maximum of 2000 feats per n:

    >>> ng = Ngrams(level='token', n_list=[1, 2], max_feats=2000)
    >>> ng.transform('this is text')
    ... {'this': 1, 'is': 1, 'text': 1, 'this is': 1, 'is text': 1}

    Notes
    -----
    Implemented by: Chris Emmery
    """

    def __init__(self, level='token', n_list=None):
        """Set parameters for N-gram extraction."""
        self.name = level+'_ngram'
        self.n_list = [2] if not n_list else n_list
        self.level = level
        self.row = 0 if level is 'token' else 2
        self.index, self.counter = 0, 0

    def __str__(self):
        """Report on feature settings."""
        return '''NGrams(level={0}, n_list={1})'''.format(self.level,
                                                          self.n_list)

    @staticmethod
    def find_ngrams(input_list, n):
        """Magic n-gram function.

        Calculate n-grams from a list of tokens/characters with added begin and
        end items. Based on the implementation by Scott Triglia.
        """
        inp = [''] * n + input_list + [''] * n
        return zip(*[inp[i:] for i in range(n)])

    def transform(self, raw, parse=None):
        """Given a document, return level-grams as Counter dict."""
        if self.level == 'char':
            needle = list(raw)
        elif self.level == 'text':
            needle = raw.split()
        elif self.level == 'token' or self.level == 'pos':
            # FIXME: parses are not handled well
            needle = parse[self.row] if parse else raw.split()
            if self.level == 'pos' and not parse:
                raise EnvironmentError("There's no POS annotation.")

        c = Counter()
        for n in self.n_list:
            c += Counter([self.level+"-"+"_".join(item) for
                          item in self.find_ngrams(needle, n)])
        return c


class WordEmbeddings(object):
    """WordEmbeddings container interacting with reach.

    Loads sparse word embeddings from pre-existing word embedding files. These
    can be either found under `datasets` in Omesa, or created using for
    example gensim, and https://github.com/mfaruqui/sparse-coding.

    Parameters
    ----------
    lang : str, optional, default 'nl'
        Language, currently only Dutch (nl) is provided in Omesa.

    mdir : str, optional, default None
        If you want to load a custom embedding file, state path.

    Attributes
    ----------
    r : class
        Spreach class initialization. More info can be found in README of
        https://github.com/stephantul/reach.
    """

    def __init__(self, lang='nl', mdir=None):
        """Load the embeddings in the proper language."""
        from reach import Spreach
        if not mdir:
            mdir = os.path.abspath(os.path.dirname(__file__)) + '/datasets/'
        if lang == 'nl':
            mdir += 'sparse-cow.txt'
        self.r = Spreach(mdir)

    def transform(self, raw, parse=None):
        """Convert sentences to tokens to vectors."""
        # FIXME: reach does an extra split if just a string is provided
        return self.r.transform(raw.split())


class FuncWords(object):
    """Extract function word frequencies.

    Computes relative frequencies of function words according to parse data,
    and adds the respective frequencies as a feature.

    Notes
    -----
    Implemented by: Chris Emmery
    Dutch functors: Ben Verhoeven
    """

    def __init__(self, lang='en'):
        """Set parameters for function word extraction."""
        self.name = 'func_words'

        if lang == 'nl':
            self.functors = {
                'VNW': 'pronouns', 'LID': 'determiners', 'VZ': 'prepositions',
                'BW': 'adverbs', 'TW': 'quantifiers', 'VG': 'conjunction'}
        else:
            raise NotImplementedError

    def transform(self, _, parse):
        """Extract frequencies for fitted function word possibilites."""
        tokens = [item[0] for item in parse if item[2].split('(')[0]
                  in self.functors]
        return Counter(tokens)


class APISent(object):
    """Sentiment features using API tools.

    Interacts with web and therefore needs urllib3. Might be _very_ slow,
    use with caution and prefrably store features.

    Parameters
    ----------
    mode : string, optional, default 'deep'
        Can be either 'deep' for Twitter-based neural sentiment (py2, boots
        local server instance), or 'nltk' for the text-processing.com API.

    Examples
    --------
    >>> sent = APISent()
    >>> sent.transform("you're gonna have a bad time")
    ... 0.030120761495050809
    >>> sent = APISent(mode='nltk')
    >>> sent.transform("you're gonna have a bad time")
    ...

    Notes
    -----
    Implemented by: Chris Emmery
    Deep sentiment: https://github.com/xiaohan2012/twitter-sent-dnn
    NLTK API: http://text-processing.com
    """

    def __init__(self, mode='deep'):
        """Load poolmanager and set API location."""
        from urllib3 import PoolManager
        self.name = 'apisent'
        self.mode = mode
        self.pool = PoolManager()

    def __str__(self):
        """String representation for APISent."""
        return '''
        feature:    {0}
        mode:       {1}
        '''.format(self.name, self.mode)

    def transform(self, raw, _):
        """Return a dictionary of feature values."""
        if self.mode == 'deep':
            jsf = json.dumps({'text': raw})
            header = {'content-type': 'application/json'}
            request = "http://localhost:6667/api"
            r = self.pool.request('POST', request, headers=header, body=jsf)
            out = {'deepsent': float(r.data.decode('utf-8'))}
        elif self.mode == 'nltk':
            qf = urlencode({'text': raw})
            request = "http://text-processing.com/api/sentiment/"
            r = self.pool.request('POST', request, body=qf)
            try:
                out = json.loads(r.data.decode('utf-8'))["probability"]
            except ValueError:
                exit("SentAPI threw unexpected response, " +
                     "probably reached rate limit.")
        return out


class DuSent(object):
    """Lexicon based sentiment features.

    Calculates four features related to sentiment: average polarity, number of
    positive, negative and neutral words. Counts based on the Duoman and
    Pattern sentiment lexicons.

    Notes
    -----
    Implemented by: Chris Emmery

    Based on code by Cynthia Van Hee, Marjan Van de Kauter, Orphée De Clercq
    """

    def __init__(self):
        """Load the sentiment lexicon."""
        self.name = 'sentiment'
        self.lexiconDict = pickle.load(
            open(__file__.split('featurizer.py')[0] +
                 '/data/sentilexicons.cpickle', 'rb'))

    def __str__(self):
        """Class string representation."""
        return '''
        feature:   %s
        ''' % (self.name)

    def calculate_sentiment(self, instance):
        """Calculate four features for the input instance.

        Instance is a list of word-pos-lemma tuples that represent a token.
        """
        polarity_score = 0.0
        token_dict = OrderedDict({
            r'SPEC\(vreemd\)': ('f', 'f'),
            r'BW\(\)': ('b', 'b'),
            r'N\(': ('n', 'n'),
            r'TWS\(\)': ('i', 'i'),
            r'ADJ\(': ('a', 'a'),
            r'WW\((od|vd).*(,prenom|,vrij)': ('a', 'v'),
            r'WW\((od|vd).*,nom': ('n', 'v'),
            r'WW\(inf,nom': ('n', 'v'),
            r'WW\(': ('v', 'v')
        })
        for parse in instance:
            try:
                word, lemma, pos, _ = parse
            except ValueError:
                exit("ERROR: DuSent relies on Frogged data!")
            for regx, param in token_dict.items():
                if re.search(regx, pos):
                    if (word, param[0]) in self.lexiconDict:
                        polarity_score += self.lexiconDict[(word, param[0])]
                    elif (lemma, param[1]) in self.lexiconDict:
                        polarity_score += self.lexiconDict[(lemma, param[1])]
                    break
                    # FIXME: reinclude the token numbers here
        return polarity_score

    def transform(self, _, parse):
        """Get the sentiment belonging to the words in the parse string."""
        return {self.name: self.calculate_sentiment(parse)}


class SimpleStats(object):
    r"""Word and token based features.

    Parameters
    ----------
    text : boolean, optional, default True
        Text-based features to be extracted, includes:

        - Total amount of flooding, and individually punctuation and
          alphanumeric stats.
        - Frequency of punctuation and number sequences.
        - Emoticon frequencies.

    token : boolean, optional, default True
        Token-based features to be extracted, includes:

        - Word lengths.
        - Number of all CAPITAL words.
        - Number Of Start Capital Words.
        - Occurence of URLs.
        - Occurence of links to pictures.
        - Occurence of links to videos.
        - Every feature listed above.

    sentence_lenth : boolean, optional, default True
        Add the sentence length as a feature.

    Examples
    --------
    All features:
    >>> SimpleStats()

    Only text features:
    >>> SimpleStats(token=False, sentence_length=False)

    Notes
    -----
    Implemented by: Chris Emmery
    Features by: Janneke van de Loo
    """

    def __init__(self, text=True, token=True, sentence_length=True):
        """Initialize all parameters to extract simple stats."""
        self.name = 'simple_stats'
        self.v = {}
        self.text, self.token, self.stl = text, token, sentence_length

    @staticmethod
    def avg(iterb):
        """Average length of iter."""
        return np.mean([len(fl) for fl, _ in iterb]) if iterb else 0

    def text_based_feats(self, raw):
        """Include features that are based on the raw text."""
        r_punc = r'[\!\?\.\,\:\;\(\)\"\'\-]'
        flood, flood_alph, flood_punc = [], [], []

        for fl in re.findall(r"((.)\2{2,})", raw):
            flood.append(len(fl[0]))
            if re.search(r'^[a-zA-Z]+$', fl[1]):
                flood_alph.append(len(fl[0]))
            if re.search(r_punc, fl[1]):
                flood_punc.append(len(fl[0]))
        # print(fl, fl_alph, fl_punc)
        av = (np.mean(flood), np.mean(flood_alph), np.mean(flood_punc))

        self.v.update({'flood_norm_len': len(flood),
                       'flood_alph_len': len(flood_alph),
                       'flood_punc_len': len(flood_punc),
                       'flood_norm_avg': av[0] if str(av[0]) != 'nan' else 0,
                       'flood_alph_avg': av[1] if str(av[1]) != 'nan' else 0,
                       'flood_punc_avg': av[2] if str(av[2]) != 'nan' else 0,
                       'num_punc': len(re.findall(r_punc + '+', raw)),
                       'num_num': len(re.findall(r'[0-9]+', raw)),
                       'num_emots': len(re.findall(r'_EMOTICON_', raw))})

    def token_based_feats(self, tokens):
        """Include features that are based on certain tokens."""
        stats = {'word_len': 0, 'cap_words': 0, 'start_cap': 0,
                 'num_urls': 0, 'num_phots': 0, 'num_vids': 0}
        r_cap = r'[A-Z\-0-9]*[A-Z][A-Z\-0-9]*$'

        for token in tokens:
            stats['word_len'] += len(token)
            stats['cap_words'] += len(''.join(re.findall(r_cap, token)))
            stats['start_cap'] += len(re.findall(r'[A-Z][a-z]', token))

            # needs token parser
            if token == '__URL__':
                stats['num_urls'] += 1
            elif token == '__PHOTO__':
                stats['num_phots'] += 1
            elif token == '__VIDEO__':
                stats['num_vids'] += 1

        self.v.update(stats)

    @staticmethod
    def avg_sent_length(sentence_indices):
        """Calculate average sentence length."""
        words_per_sent = Counter(sentence_indices)
        return np.mean([val for _, val in words_per_sent.items()])

    def transform(self, raw, parse):
        """Transform given instance into simple text features."""
        if self.text:
            self.text_based_feats(raw)
        try:
            if self.token:
                assert parse
                self.token_based_feats([p[0] for p in parse])
            if self.stl:
                assert parse
                self.avg_sent_length([p[3] for p in parse])
        except AssertionError:
            exit("SimpleStats - No parses were found to extract token or " +
                 "sentence features from. Please provide or disable features.")
        return self.v


class Readability(object):
    """Get readability-related features.

    Notes
    -----
    Implemented by: Chris Emmery
    Attributes by: Tom De Smedt
    """

    def __init__(self):
        """Initialize empty class variables."""
        self.name = 'readability'
        self.diacritics = \
            u"àáâãäåąāæçćčςďèéêëēěęģìíîïīłįķļľņñňńйðòóôõöøþřšťùúûüůųýÿўžż"
        self.punctuation = ".,;:!?()[]{}`''\"@#$^&*+-|=~_"
        self.flooding = re.compile(r"((.)\2{2,})", re.I)  # ooo, xxx, !!!, ...
        self.emoticons = set((
            '*)', '*-)', '8)', '8-)', '8-D', ":'''(", ":'(", ':(', ':)',
            ':-(', ':-)', ':-.', ':-/', ':-<', ':-D', ':-O', ':-P', ':-S',
            ':-[', ':-b', ':-c', ':-o', ':-p', ':-s', ':-|', ':/', ':3', ':>',
            ':D', ':O', ':P', ':S', ':[', ':\\', ':]', ':^)', ':b', ':c',
            ':c)', ':o', ':o)', ':p', ':s', ':{', ':|', ':}', ";'(", ';)',
            ';-)', ';-]', ';D', ';]', ';^)', '<3', '=(', '=)', '=-D', '=/',
            '=D', '=]', '>.>', '>:)', '>:/', '>:D', '>:P', '>:[', '>:\\',
            '>:o', '>;]', 'X-D', 'XD', 'o.O', 'o_O', 'x-D', 'xD', u'\xb0O\xb0',
            u'\xb0o\xb0', u'\u2665', u'\u2764', '^_^', '-_-'
        ))
        self.emoji = set((
            u'\U0001f44c', u'\U0001f44d', u'\U0001f47f', u'\U0001f495',
            u'\U0001f499', u'\U0001f49a', u'\U0001f49b', u'\U0001f49c',
            u'\U0001f600', u'\U0001f601', u'\U0001f602', u'\U0001f603',
            u'\U0001f604', u'\U0001f605', u'\U0001f606', u'\U0001f607',
            u'\U0001f608', u'\U0001f60a', u'\U0001f60b', u'\U0001f60c',
            u'\U0001f60d', u'\U0001f60e', u'\U0001f60f', u'\U0001f610',
            u'\U0001f612', u'\U0001f613', u'\U0001f614', u'\U0001f615',
            u'\U0001f61b', u'\U0001f61c', u'\U0001f61d', u'\U0001f61e',
            u'\U0001f61f', u'\U0001f620', u'\U0001f621', u'\U0001f622',
            u'\U0001f625', u'\U0001f626', u'\U0001f627', u'\U0001f629',
            u'\U0001f62a', u'\U0001f62b', u'\U0001f62c', u'\U0001f62d',
            u'\U0001f62e', u'\U0001f62f', u'\U0001f633', u'\U0001f636',
            u'\U0001f63b', u'\U0001f63f', u'\U0001f640', u'\u2764\ufe0f',
            u'\u263a', u'\ud83d', u'\ude09'
        ))
        self.url = re.compile(r"https?://[^\s]+")
        self.ref = re.compile(r"@[a-z0-9_./]+", flags=re.I)

    def transform(self, raw, _):
        """Add each metric to the feature vector."""
        # TODO: add stuff here
        raw = raw + _
        return NotImplementedError



"""Experiment, storage and interpretation module for machine learning research.

A small framework for reproducible machine learning research that largely
builds on top of scikit-learn. Its goal is to make common research procedures
fully automated, optimized, and well recorded.
"""

from omesa import experiment
from omesa import containers
from omesa import featurizer

__author__ = 'Chris Emmery'
__contrb__ = ('Akos Kadar', 'Mike Kestemont', 'Florian Kunneman',
              'Janneke van de Loo', 'Ben Verhoeven')
__license__ = 'GPLv3'
__version__ = '0.3.0a0'

__all__ = ['experiment', 'containers', 'featurizer']



"""
Python and numpy serialization module.

Edit by Chris Emmery, for https://cmry.github.io/notes/_serialize.

All credits go to:

Copyright (c) 2013, Christopher R. Wagner

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation files
(the "Software"), to deal in the Software without restriction,
including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software,
and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

# pylint:        disable=W0612,W0122,R0204

from collections import namedtuple, OrderedDict
from types import GeneratorType
from multiprocessing import Pool
import json
import sys
import numpy as np

POOL = False


class Dummy(object):
    """Dummy class to reinitialize."""

    def __init__(self):
        """Empty init."""
        pass


def mod_load(mod, name):
    """Module loader."""
    try:
        getattr(sys.modules[mod], name)
    except KeyError:
        exec("from " + mod + " import " + name)
    return getattr(sys.modules[mod], name)


def isnamedtuple(obj):
    """Heuristic check if an object is a namedtuple."""
    return isinstance(obj, tuple) \
        and hasattr(obj, "_fields") \
        and hasattr(obj, "_asdict") \
        and callable(obj._asdict)


def _serialize(data):
    if data is None or isinstance(data, (bool, int, float, str)):
        return data
    if isinstance(data, list):
        return [_serialize(val) for val in data]
    if isinstance(data, OrderedDict):
        return {"py/collections.OrderedDict":
                [[_serialize(k), _serialize(v)] for k, v in data.items()]}
    if isnamedtuple(data):
        return {"py/collections.namedtuple": {
            "type":   type(data).__name__,
            "fields": list(data._fields),
            "values": [_serialize(getattr(data, f)) for f in data._fields]}}
    # --- custom ---
    if isinstance(data, type):
        return {"py/numpy.type": data.__name__}
    if isinstance(data, np.integer):
        return {"py/numpy.int": int(data)}
    if isinstance(data, np.float):
        return {"py/numpy.float": data.hex()}
    # -------------
    if isinstance(data, dict):
        if all(isinstance(k, str) for k in data):
            return {k: _serialize(v) for k, v in data.items()}
        return {"py/dict": [[_serialize(k), _serialize(v)]
                            for k, v in data.items()]}
    if isinstance(data, tuple):
        return {"py/tuple": [_serialize(val) for val in data]}
    if isinstance(data, set):
        return {"py/set": [_serialize(val) for val in data]}
    if isinstance(data, np.ndarray):
        return {"py/numpy.ndarray": {
            "values": data.tolist(),
            "dtype":  str(data.dtype)}}
    # --- custom ---
    if isinstance(data, GeneratorType):
        return {'py/generator': str(data)}
    if not isinstance(data, type) and hasattr(data, '__module__'):
        return {'py/class': {'name': data.__class__.__name__,
                             'mod': data.__module__,
                             'attr': _serialize(data.__dict__)}}
    if '_csv.reader' in str(type(data)):
        return ''
    raise TypeError("Type %s not data-serializable" % type(data))


def _restore(dct):
    # --- custom ---
    if "py/numpy.type" in dct:
        return np.dtype(dct["py/numpy.type"]).type
    elif "py/numpy.int" in dct:
        return np.int32(dct["py/numpy.int"])
    elif "py/numpy.float" in dct:
        return np.float64.fromhex(dct["py/numpy.float"])
    # -------------
    elif "py/dict" in dct:
        return dict(dct["py/dict"])
    elif "py/tuple" in dct:
        return tuple(dct["py/tuple"])
    elif "py/set" in dct:
        return set(dct["py/set"])
    elif "py/collections.namedtuple" in dct:
        data = dct["py/collections.namedtuple"]
        return namedtuple(data["type"], data["fields"])(*data["values"])
    elif "py/numpy.ndarray" in dct:
        data = dct["py/numpy.ndarray"]
        return np.array(data["values"], dtype=data["dtype"])
    elif "py/collections.OrderedDict" in dct:
        return OrderedDict(dct["py/collections.OrderedDict"])
    # --- custom ---
    elif "py/generator" in dct:
        return []
    elif "py/class" in dct:
        obj = dct["py/class"]
        cls_ = mod_load(obj['mod'], obj['name'])
        class_init = Dummy()
        class_init.__class__ = cls_
        for k, v in _restore(obj['attr']).items():
            setattr(class_init, k, v)
        return class_init
    return dct


def pool(func, data):
    """Multi-thread serialization."""
    global POOL
    if not POOL:
        POOL = True
        p = Pool(processes=None)
        out = p.map(func, data)
        p.close()
        p.join()
        del p
        POOL = False
    else:
        out = data
    return out


def encode(data, fp=False):
    """Python object to file or string."""
    if fp:
        return json.dump(_serialize(data), fp)
    else:
        return json.dumps(_serialize(data))


def decode(fp):
    """File, String, or Dict to python object."""
    try:
        return json.load(fp, object_hook=_restore)
    except (AttributeError, ValueError):
        pass
    try:
        return json.loads(fp, object_hook=_restore)
    except (TypeError, ValueError):
        pass
    return json.loads(json.dumps(fp), object_hook=_restore)



"""Experiment wrapper code.
"""

# License:      GPLv3

from time import time

import numpy as np

from sklearn import metrics
from sklearn.cross_validation import cross_val_predict
from sklearn.cross_validation import train_test_split

from .logger import _Logger as Logger
from .components import Vectorizer, Optimizer
from .containers import Pipeline


class Experiment(object):
    """Full experiment wrapper.

    Calls several sklearn modules in the Pipeline class and reports on the
    classifier performance. For this, the class uses a configuration
    dictionary. The full list of options for this is listed under attributes.

    Parameters
    ----------
    project : string
        The project name functions as a hook to for example call the best
        performing set of parameters out of a series of experiments on the same
        data.

    name : string
        Same as the above. This will function as a hook to save your model,
        features and Omesa config under one name.

    train_data : list of, or single iterator
        Example: [CSV("/somedir/train.csv", label=1, text=2),
                  CSV("/somedir/train2.csv", label=3, text=5]
        The data on which the experiment will train. If the location of a .csv
        is provided, it will open these up and create an iterator for you.
        Alternatively, you can provide your own iterators or iterable
        structures providing instances of the data. If only training data is
        provided, the experiment will evaluate in a tenfold setting by default.

    test_data : list of, or single iterator
        Example: [CSV("/somedir/test.csv", label=1, text=2)]
        This works similar to the train_data. However, when a test set is
        provided, the performance of the model will be measured on this test
        data only. Omesa will dump a classification report for you.

    test_proportion : float
        Example: 0.3
        As opposed to a test FILE, one can also provide a test proportion,
        after which a certain amount of instances will be held out from the
        training data to test on.

    features : list of featurizer classes
        Example: [Ngrams()]
        These can be features imported from omesa.featurizer or can be any
        class you create your self. As long as it adheres to a fit / transform
        structure and returns a feature dictionary per instance as can be
        provided to, for example, the sklearn FeatureHasher.

    backbone : class from backbone
        Example: Spacy()
        The backbone is used as an all-round NLP toolkit for tagging, parsing
        and in general annotating the text that is provided to the experiment.
        If you wish to utilize features that need for example tokens, lemmas or
        POS tags, they can be parsed during loading. Please be advised that
        it's more convenient to do this yourself beforehand.

    classifier : class with sklearn API
        Example: GaussianNB()
        Used to switch the classifier used in the experiment. By default, an
        SVM with low parameter settings is used if you do NOT want to use grid
        search. In any other case, you can provide other sklearn classifiers
        that can be set to output probabilities.

    save : tuple of strings
        Example: ("log", model", "db", "man", "json", "pickle")
        Save the output of the log, or dump the entire model with its
        classification method and pipeline wrapper for new data instances.
    """

    def __init__(self, **kwargs):
        """Set all relevant classes, run experiment (currently)."""
        self.conf = kwargs
        self.log = Logger(self.conf['name'])
        self.vec = Vectorizer(self.conf)
        self.opt = Optimizer(self.conf)
        self.clf = None
        self.clf_unfit = None
        self.res = {}
        self._run(self.conf)

    def save(self):
        """Save desired Experiment data."""
        if self.conf.get('save'):
            if 'log' in self.conf['save']:
                self.log.save()
            if 'features' in self.conf['save']:
                self.log.echo(" Feature saving has not been implemented yet!")
            if 'model' in self.conf['save']:
                Pipeline(self).save()

    def _run_proportions(self, sets, av, seed, pr=None):
        """Repeats run and proportionally increases the amount of data."""
        X, Xi, y, yi = sets
        pr = self.conf.get('proportions', pr)
        self.res['prop'], tmp, conf = {}, self.opt, self.conf
        for i in range(1, pr):
            self.opt, prop = Optimizer(conf), (1 / pr) * (pr - i)
            self.log.slice((1 - prop, ))
            Xp, _, yp, _ = train_test_split(X, y, test_size=prop, stratify=y)
            clff = self.opt.choose_classifier(Xp, yp, seed).fit(Xp, yp)
            tres = cross_val_predict(clff, Xp, yp, cv=5, n_jobs=-1)
            tscore = metrics.f1_score(yp, tres, average=av)
            score = metrics.f1_score(yi, clff.predict(Xi), average=av)
            print("\n Result: {0}".format(score))
            self.res['prop'].update(
                {1 - prop: {'train': tscore, 'test': score}})
        self.opt = tmp

    def _run(self, conf):
        """Split data, fit, transfrom features, tf*idf, svd, report."""
        t1 = time()
        seed = 42
        np.random.RandomState(seed)

        # report features
        self.log.head(conf['features'], conf['name'], seed)

        # stream data to sparse features

        X, y = self.vec.fit_transform(conf['train_data'])
        # split off test data
        if not conf.get('test_data'):
            X, Xi, y, yi = train_test_split(
                X, y, test_size=conf.get('test_proportion', 0.1), stratify=y)

        # grid search and fit best model choice
        self.clf = self.opt.choose_classifier(X, y, seed)
        print("\n Training model...")
        self.clf.fit(X, y)
        print(" done!")

        av = 'binary' if len(set(y)) == 2 else 'micro'
        labs = self.vec.encoder.classes_
        self.log.data('sparse', 'train', X)
        # if user wants to report more than best score, do another CV on train
        if conf.get('detailed_train', True):
            res = cross_val_predict(self.clf, X, y, cv=5, n_jobs=-1)
            self.res['train'] = self.log.report('train', y, res, av, metrics,
                                                labs)

        # report performance
        if conf.get('test_data'):
            Xi, yi = self.vec.transform(conf['test_data'])

        self.log.data('sparse', 'test', Xi, dump=True)
        res = self.clf.predict(Xi)
        self.res['test'] = self.log.report('test', yi, res, av, metrics, labs)

        if conf.get('proportions'):
            self._run_proportions((X, Xi, y, yi), av, seed)

        print("\n # ------------------------------------------ \n")
        t2 = time()
        dur = round(t2 - t1, 1)
        self.res['dur'] = dur
        print("\n Experiment took {0} seconds".format(dur))

        self.save()
        print("\n" + '-' * 10, "\n")



"""Setup file."""

from setuptools import setup
from setuptools import find_packages


setup(name='omesa',
      version='0.3.2.alpha',
      description='A framework for reproducible machine learning research',
      author='Chris Emmery',
      author_email='chris.emmery@uantwerpen.be',
      url='https://github.com/cmry/omesa',
      license='GPLv3',
      packages=find_packages(exclude=['tests']),
      install_requires=['blitzdb>=0.2.12',
                        'bottle>=0.12.9',
                        'colorlover>=0.2.1',
                        'lime>=0.1.1.7',
                        'plotly>=1.12.0',
                        'reach>=0.0.4',
                        'scikit-learn>=0.17'],
      zip_safe=True)
